@@ -424,7 +424,8 @@ class SparseMatrixRecommender:
     # ------------------------------------------------------------------
     # Recommend by profile
     # ------------------------------------------------------------------
-    def recommend_by_profile(self, profile, nrecs=10, normalize=True, ignore_unknown=False, vector_result: bool = False):
+    def recommend_by_profile(self, profile, nrecs=10, normalize=True, ignore_unknown=False,
+                             vector_result: bool = False):
         """Recommend by profile.
 
         :type profile: str|list|dict
@@ -599,6 +600,155 @@ class SparseMatrixRecommender:
 
         return self
 
+    # ------------------------------------------------------------------
+    # Filter by profile
+    # ------------------------------------------------------------------
+    def filter_by_profile(self,
+                          profile,
+                          filter_type="intersection",
+                          ignore_unknown=False):
+        """
+        Filter by profile
+        -----------------
+        The result is a vector of scored items that is assigned to \code{smrObj$Value}.
+        If filter_type is "union" each item that has at least one of the tags in profile is in the result.
+        (Essentially, that is the same as taking all non-zero score recommendations by profile.)
+        If filter_type is "intersection" each item in the result has all tags in profile.
+
+        :param profile: A profile specification used to filter with.
+        :param filter_type: The type of filtering one of "union" or "intersection".
+        :param ignore_unknown:
+        :return self: The object itself or None. The result is stored in self._value.
+        """
+
+        # Make scored tags vector
+        if isinstance(profile, str):
+            vec = self.to_profile_vector([profile], ignore_unknown=ignore_unknown).take_value()
+        elif isinstance(profile, dict) or isinstance(profile, list):
+            vec = self.to_profile_vector(profile, ignore_unknown=ignore_unknown).take_value()
+        elif is_s_sparse_matrix(profile):
+            vec = profile
+        else:
+            raise TypeError("The first argument is expected to be a list of tags or a dictionary of scored tags.")
+
+        # Unitize
+        profileVec = vec.unitize()
+
+        # Find the items corresponding to the profile and type spec
+        if isinstance(filter_type, str) and filter_type.lower() == "union":
+
+            sVec = self.recommend_by_profile(profileVec, nrecs=None).take_value()
+            sVec = sVec.keys()
+
+        elif isinstance(filter_type, str) and filter_type.lower() == "intersection":
+
+            sVec = self.take_M().unitize().dot(profileVec).row_sums_dict()
+
+            n = profileVec.column_sums()[0]
+            sVec = [k for (k, v) in sVec.items() if v >= n]
+        else:
+            raise TypeError("The argument filter_type is expected to be one of \"union\" or \"intersection\".")
+
+        # Result
+        self.set_value(sVec)
+        return self
+
+    # ------------------------------------------------------------------
+    # Retrieve by query elements
+    # ------------------------------------------------------------------
+    def retrieve_by_query_elements(self,
+                                   should=[],
+                                   must=[],
+                                   must_not=[],
+                                   must_type="intersection",
+                                   must_not_type="union",
+                                   ignore_unknown=False):
+        """
+        Retrieve by query elements
+        --------------------------
+        Applies a profile filter to the rows of the recommendation matrix.
+ 
+        :param should: A profile specification used to recommend with.
+        :param must: A profile specification used to filter with. The items in the result must have the tags in the given list.
+        :param must_not: A profile specification used to filter with. The items in the result must not have the tags in given list
+        :param must_type: The type of filtering with the must tags; one of "union" or "intersection".
+        :param must_not_type: The type of filtering with the must not tags; one of "union" or "intersection".
+        :param ignore_unknown: Should the unknown tags be ignored or not?
+        :return self:  The object itself or None. The result is stored in self._value.
+        """
+
+        if not is_str_list(should):
+            raise TypeError("The first argument, should, is expected to be a list of tags.")
+
+        if not is_str_list(must):
+            raise TypeError("The second argument, must, is expected to be a list of tags.")
+
+        if not is_str_list(must_not):
+            raise TypeError("The third argument, must_not, is expected to be a list of tags.")
+
+        if len(should) + len(must) + len(must_not) == 0:
+            warnings.warn("All query elements are empty.")
+            return self
+
+        # Should
+        if len(should) > 0 and len(must) > 0:
+            # Both should and must are present
+            pVecShould = self.to_profile_vector(should, ignore_unknown=ignore_unknown).take_value()
+            pVecMust = self.to_profile_vector(must, ignore_unknown=ignore_unknown).take_value()
+
+            shouldItems = self.recommend_by_profile(pVecShould.add(pVecMust),
+                                                    nrecs=None,
+                                                    ignore_unknown=ignore_unknown).take_value()
+
+        elif len(should) > 0 and len(must) == 0 and len(must_not) == 0:
+            # Only should is not empty
+
+            pVecShould = self.to_profile_vector(should, ignore_unknown=ignore_unknown).take_value()
+            self.recommend_by_profile(pVecShould,
+                                      nrecs=None,
+                                      ignore_unknown=ignore_unknown)
+
+            return self
+
+        else:
+            shouldItems = dict.fromkeys(self.take_M().row_names(), 1)
+
+        res = shouldItems
+
+        # Must
+        if len(must) > 0:
+            mustItems = self.filter_by_profile(must,
+                                               filter_type=must_type,
+                                               ignore_unknown=ignore_unknown).take_value()
+
+            if len(mustItems) == 0:
+                warnings.warn("No items were obtained by querying with the must tags.")
+
+        else:
+            mustItems = []
+
+        if len(mustItems) > 0:
+            res = set.intersection(set(res), set(mustItems))
+
+        # Must not
+        if len(must_not) > 0:
+            mustNotItems = self.filter_by_profile(must_not,
+                                                  filter_type=must_not_type,
+                                                  ignore_unknown=ignore_unknown).take_value()
+
+            if len(mustNotItems) == 0:
+                warnings.warn("No items were obtained by querying with the must not tags.")
+
+        else:
+            mustNotItems = []
+
+        if len(mustNotItems) > 0:
+            res = set.difference(set(res), set(mustNotItems))
+
+        # Result
+        self.set_value(dict.fromkeys(res, 1))
+        return self
+
     # ------------------------------------------------------------------
     # Join across
     # ------------------------------------------------------------------
@@ -637,11 +787,57 @@ class SparseMatrixRecommender:
 
         # Left join recommendations with the given data frame
         # Using .join does not work because .join uses indexes.
-        # Hence using merge.
+        # Hence, using merge.
         self.set_value(dfRecs.merge(data, on=mon, how="left"))
 
         return self
 
+    # ------------------------------------------------------------------
+    # Remove tag types
+    # ------------------------------------------------------------------
+    def remove_tag_types(self,
+                         tag_types,
+                         warn: bool = True):
+        """Remove tag types.
+
+        :type tag_types: str|list
+        :param tag_types: A list of tag types to be removed from the SMR object.
+
+        :type warn: bool
+        :param warn: Should warning messages be issued or not?
+
+        :rtype SparseMatrixRecommender
+        :return self: The object itself or None.
+        """
+
+        # There are several ways to do this:
+        # 1. Work with newSMR$TagTypeRanges, take the indices corresponding to tag types not to be removed.
+        # 2. Construct a metadata matrix by taking sub-matrices of the tag types not to be removed.
+
+        removeTagTypes = tag_types
+
+        # Process tag_types
+        if isinstance(removeTagTypes, str):
+            removeTagTypes = [removeTagTypes, ]
+
+        if not is_str_list(removeTagTypes):
+            raise ValueError("The argument tag_types is expected to be a string or a list of strings.")
+
+        tagTypes = list(self.take_matrices().keys())
+        tagTypesKnown = list(set.intersection(set(tagTypes), set(removeTagTypes)))
+
+        if len(tagTypesKnown) == 0:
+            raise ValueError("None of the specified tag types is a known tag type in the recommender object.")
+
+        if len(tagTypesKnown) < len(tagTypes):
+            warnings.warn("Some tags are not known in the recommender.")
+
+        tagTypes = set.difference(set(tagTypes), set(removeTagTypes))
+
+        res = SparseMatrixRecommender({t: self.take_matrices()[t] for t in tagTypes})
+
+        return res
+
     # ------------------------------------------------------------------
     # Classify by profile
     # ------------------------------------------------------------------
@@ -829,6 +1025,146 @@ class SparseMatrixRecommender:
         # Create recommender
         return SparseMatrixRecommender().create_from_matrices(matrices=smats)
 
+    # ------------------------------------------------------------------
+    # SMR join
+    # ------------------------------------------------------------------
+    def join(self, smr2, join_type="left", **kwargs):
+
+        allRowNames = self.take_M().row_names()
+        if join_type != "same":
+
+            if join_type in set(["outer", "union"]):
+
+                allRowNames = list(set.union(set(self.take_M().row_names()),
+                                             set(smr2.take_M().row_names())))
+
+            elif join_type == "inner":
+
+                allRowNames = list(set.intersection(set(self.take_M().row_names()),
+                                                    set(smr2.take_M().row_names())))
+
+            elif join_type == "left":
+
+                allRowNames = self.take_M().row_names()
+
+            else:
+                raise ValueError(
+                    "The second argument is expected to be one of \"same\", \"outer\", \"inner\", \"left\".")
+
+        aSMats1 = self.take_matrices()
+        aSMats2 = smr2.take_matrices()
+
+        commonTags = set.intersection(set(aSMats1.keys()), set(aSMats2.keys()))
+        if len(commonTags) > 0:
+            warnings.warn(
+                "The tag types " + str(commonTags) + " are also in the SMR argument, hence will be dropped.")
+
+        if join_type != "same":
+            aSMats1 = {k: v.impose_row_names(allRowNames) for (k, v) in aSMats1.items()}
+            aSMats2 = {k: v.impose_row_names(allRowNames) for (k, v) in aSMats2.items()}
+
+        # Not using this nice way, since it is "only" for >=3.9
+        # matrices = aSMats1 | aSMats2
+        matrices = {**aSMats1, **aSMats2}
+
+        return SparseMatrixRecommender(matrices)
+
+    # ------------------------------------------------------------------
+    # Annex matrices
+    # ------------------------------------------------------------------
+    def annex_sub_matrices(self, mats):
+        """Enhance recommender matrix with nearest neighbors sub-matrix.
+
+        :tupe mats: dict
+        :param mats: A dictionary of matrices to be annexed.
+
+        :rtype: SparseMatrixRecommender
+        :return A sparse matrix recommender
+        """
+
+        if not is_smat_dict(mats):
+            raise ValueError("The second argument, mats, is expected to be a dictionary of SSparseMatrix.")
+
+        smr2 = SparseMatrixRecommender(mats)
+        return self.join(smr2=smr2)
+
+    # ------------------------------------------------------------------
+    # Annex matrix
+    # ------------------------------------------------------------------
+    # I consider the definition of this function to be redundant.
+    # def annex_sub_matrix(self, new_tag_type, new_mat):
+    #    return self.annex_sub_matrices({new_tag_type: new_mat})
+
+    # ------------------------------------------------------------------
+    # To metadata recommender
+    # ------------------------------------------------------------------
+    def enhance_with_nearest_neighbors(self,
+                                       tag_types=None,
+                                       number_of_nearest_neighbors=20,
+                                       nearest_neighbors_tag_type="neighbor",
+                                       batch_size=None):
+        """Enhance recommender matrix with nearest neighbors sub-matrix.
+
+        :tupe tag_types: str|list|None
+        :param tag_types: Tag types to use make compute the similarities.
+
+        :type number_of_nearest_neighbors: Int
+        :param number_of_nearest_neighbors: Number of nearest neighbors per item.
+
+        :type nearest_neighbors_tag_type: str
+        :param nearest_neighbors_tag_type: Tag type of the nearest neighbor sub-matrix.
+
+        :type batch_size: int|None
+        :param batch_size: Number of nearest neighbors per item.
+
+        :rtype: SparseMatrixRecommender
+        :return A sparse matrix recommender
+        """
+
+        tagTypes = tag_types
+        # Filter tag types
+        if isinstance(tag_types, str):
+            tagTypes = [tagTypes, ]
+
+        if tag_types is None:
+            tagTypes = list(self.take_matrices().keys())
+
+        if not is_str_list(tagTypes):
+            raise ValueError("The argument tag_types is expected to be a string, a list of string, or None.")
+
+        removeTagTypes = set.difference(set(self.take_matrices().keys()), set(tagTypes))
+
+        if len(removeTagTypes) > 0:
+            smrRes = self.remove_tag_types(tag_types=removeTagTypes)
+        else:
+            smrRes = self
+
+        # Similarities computations
+        # This implementation is far from optimal.
+        dfNNs = None
+        for itemID in smrRes.take_M().row_names():
+
+            aRecs = smrRes.recommend(history=itemID,
+                                     nrecs=number_of_nearest_neighbors,
+                                     remove_history=False,
+                                     normalize=True).take_value()
+
+            dfRecs = pandas.DataFrame(
+                {"SearchID": itemID, "NearestNeighborID": list(aRecs.keys()), "Score": list(aRecs.values())})
+
+            if dfNNs is None:
+                dfNNs = dfRecs
+            else:
+                dfNNs = pandas.concat([dfNNs, dfRecs], ignore_index=True)
+
+        # Annex sub-matrix
+        smatNNs = cross_tabulate(dfNNs, "SearchID", "NearestNeighborID", "Score")
+
+        smrRes = self.annex_sub_matrices(mats={nearest_neighbors_tag_type: smatNNs})
+
+        # Result
+        return smrRes
+
     # ------------------------------------------------------------------
     # To dictionary form
     # ------------------------------------------------------------------
